// Grid-1.cc
//
// Implementation for methods of Grid<1>
//

#include "splines/Grid.h"
#include "splines/GridSearchAnswer.h"
#include "splines/DoubConv.h"
#include <iostream>
#include <sstream>
#include <string>
#include <cmath>
#include <algorithm>

//#define TRACE_G0
//#define TRACE_G1
//#define TRACE_G2

namespace splines {

Grid<1>::Grid (const Grid<1> & g) :
        upperBoundary    (g.upperBoundary),
        penultimatePoint (g.penultimatePoint),
        penultimateK     (g.penultimateK),
        lastIndex        (g.lastIndex),
        pointCount       (g.pointCount),
        x(g.x),
        h(g.h),
        k(g.k),
        index(g.index)
{}

void
Grid<1>::swap (Grid<1> & g) {
  std::swap (upperBoundary,    g.upperBoundary);
  std::swap (penultimatePoint, g.penultimatePoint);
  std::swap (penultimateK,     g.penultimateK);
  std::swap (lastIndex,        g.lastIndex);
  std::swap (pointCount,       g.pointCount);
  x.swap(g.x);
  k.swap(g.k);
  h.swap(g.h);
  index.swap(g.index);
}

void
Grid<1>::push_back (double p) {
				#ifdef TRACE_G2
  				std::cout << "push_back(" << p 
					  << "):  before -- \n";
  				put (std::cout);
				#endif
  if (lastIndex >= 0) {
    if (p <= upperBoundary) {
      throw " Grid += a point at p requires p > the upperBoundary";
    }
    double kTemp = 1.0/(p-upperBoundary);
    x.push_back     ( upperBoundary );
    h.push_back     (p-upperBoundary);
    k.push_back     ( kTemp );
    index.push_back ( lastIndex );
    penultimatePoint = upperBoundary;
    penultimateK = kTemp;
  }
  upperBoundary = p;
  lastIndex++;
  pointCount++;
				#ifdef TRACE_G2
  				std::cout << "                            "
				          << "after  -- \n";
  				put (std::cout);
				#endif
}

Grid<1> &
Grid<1>::operator += (const Grid<1> & g) {
  if (g.lastIndex == -1) { return *this; }
  if (lastIndex   == -1) { return operator= ( g ); }

  // Do checking for order early, for exception safety
  if (g.x.empty()) {
    if (g.upperBoundary <= upperBoundary) {
      throw Grid_append_order_exception (g.upperBoundary, upperBoundary);
    }
  } else {
    if (g.x[0] <= upperBoundary) {
      throw Grid_append_order_exception (g.x[0], upperBoundary);
    }
  }

  x.push_back ( upperBoundary );
  index.push_back (lastIndex);
  ++lastIndex;
  if (g.x.empty()) {
    penultimatePoint = upperBoundary;
    penultimateK     = 1.0/(g.upperBoundary-upperBoundary);
    h.push_back ( g.upperBoundary-upperBoundary );
    k.push_back ( penultimateK );
  } else {
    penultimatePoint = g.penultimatePoint;
    penultimateK =     g.penultimateK;
    h.push_back ( g.x[0]-upperBoundary );
    k.push_back ( 1.0/(g.x[0]-upperBoundary) );
    copy ( g.x.begin(),     g.x.end(),     back_inserter(x) );
    copy ( g.k.begin(),     g.k.end(),     back_inserter(k) );
    copy ( g.h.begin(),     g.h.end(),     back_inserter(h) );
    std::vector<int>::const_iterator end (g.index.end());
    for (std::vector<int>::const_iterator i = g.index.begin(); i != end; ++i){
      index.push_back( *i + lastIndex );
    }
    lastIndex += g.lastIndex;
  }
  upperBoundary = g.upperBoundary;
  pointCount += g.pointCount;
  return *this;
}

Grid<1>::Grid ( double lowerLimit,
                double upperLimit,
                unsigned int nGridPoints ) {
  if (upperLimit <= lowerLimit) {
    throw Grid_limits_order_exception (lowerLimit, upperLimit);
  }
  if (nGridPoints == 0) {
    lastIndex = -1; // empty grid
    return;
  }
  if (nGridPoints == 1) {
    lastIndex = 0;
    upperBoundary = lowerLimit;
    return;
  }
  int N = nGridPoints;
  double gridX = (upperLimit-lowerLimit);
  double gridK = N/gridX;
  double gridH = gridX/N;
  x.push_back (lowerLimit);
  h.push_back (gridH);
  k.push_back (gridK);
  index.push_back(0);
  penultimatePoint = lowerLimit + (N-2)*gridH;
  penultimateK     = gridK;
  upperBoundary    = lowerLimit + (N-1)*gridH;
  lastIndex = N-1;
  pointCount = nGridPoints;
}

bool 
Grid<1>::search (double p, GridSearchAnswer & answer) const {
				#ifdef TRACE_G1
					static int ntraces = 0;
					// if (ntraces<10) { ntraces++;
					if (1) { ntraces++;
					std::cout << "Grid search: p = " 
					          << p << "\n"
					          << " penultimatePoint = " 
						  << penultimatePoint << "\n";
					std::cout << "ntraces = " << ntraces 
					          << "\n";
					}
				#endif
  if ( p < penultimatePoint ) {
    std::vector<double>::const_iterator i = 
				std::upper_bound ( x.begin(), x.end(), p );
    int m = i-x.begin() - 1;
    double distance;
    double ndist;
    int nndist;
				#ifdef TRACE_G1
				std::cout << "m = " << m << "\n";
				#endif
    if ( m < 0 ) {
      distance = (p - x[0]) * k[0];
      int nndist = 0;
      answer.h = h[0];
      answer.index = 0;
      answer.x     = distance;
				#ifdef TRACE_G1
				std::cout << "distance = (p - x[0]) * k[0] = " 
				          << distance << " h[0] = "
					  << answer.h << " nndist = 0 \n";
				#endif
    } else {
      distance = (p - x[m]) * k[m];
      ndist = std::floor(distance);
      nndist = static_cast<int> (ndist);
      answer.h = h[m];
      answer.index = index[m] + nndist;
      answer.x     = distance - ndist;
				#ifdef TRACE_G1
				std::cout << "distance = (p - x[m]) * k[m] = " 
				          << distance << " h[0] = "
					  << answer.h << " nndist = " 
					  << nndist << "\n";
				#endif
    }
				#ifdef TRACE_G1
				std::cout << "index[m] = " << index[m]
					  << " distance = " << distance << "\n"
					  << "answer.index = " << answer.index
					  << " .x = " << answer.x 
					  << " .h = " << answer.h 
					  << " return = " << (distance >= 0)
					  << "\n"; 
				#endif
    return (distance >= 0);
  } else {  // (p >= penultimatePoint) 
				#ifdef TRACE_G1
				// if (ntraces<11) { 
				if (1) { 
				std::cout << "p is beyone penultimate point\n"
				<< "penultimatePoint = " << penultimatePoint
				<< "penultimateK = " << penultimateK
				<< "\ndistance = " << (p - penultimatePoint)*penultimateK
				<< " index = " << lastIndex-1
				<< " x = " << (p - penultimatePoint)*penultimateK 
				<< " h = " << 1.0/penultimateK << "\n";
	}
				#endif
    answer.index = lastIndex-1;
    answer.x     = (p - penultimatePoint)*penultimateK;
    answer.h     = 1.0/penultimateK;
				#ifdef TRACE_G1
				std::cout << "lastIndex = " << lastIndex-1
					  << " penultimatePoint = " << penultimatePoint 
					  << " penultimateK = " << penultimateK
					  << "\n"
					  << "answer.index = " << answer.index
					  << " .x = " << answer.x 
					  << " .h = " << answer.h 
					  << " return = " << (answer.x <= 1.0)
					  << "\n"; 
				#endif
    return ( answer.x <= 1.0 );
  }
}

double
Grid<1>::operator[] (unsigned int p) const {
  unsigned int prev = 0;
  unsigned int end = index.size();
  unsigned int i;

  assert (p <= lastIndex);
  if ( p >= lastIndex ) return upperBoundary;
                    
  // We want to find the highest value i such that index[i] <= ind.
  // $$$ See technical note G-1
  i = (p >= index[prev]) ? prev+1 : 1;
  while ( (i < end) && index[i] <= p ) { ++i; }
  --i;
  prev = i;

  // Find x at point p, knowing that we are in region i.
  // $$$ See technical note G-2
   return (x[i] + (p-index[i])/k[i]);
}

std::vector<double>
Grid<1>::points () const {
  std::vector<double> v;
  v.reserve(pointCount);

  // All regions up to but not including the last one
  unsigned int end = index.size()-1;
  for (unsigned int i = 0; i < end; ++i) {
				#ifdef TRACE_G0
				std::cout<< "Region " << i << "\n";
				#endif
    unsigned int last_p = index[i+1];
    for (unsigned int p = index[i]; p < last_p; ++p) {
				#ifdef TRACE_G0
				std::cout << "p " << p << " index[i] " 
				          << index[i]
        				  <<  " k[i] " << k[i] 
					  <<  " x[i] " << x[i] 
					  << " push_back(" 
					  << x[i]+(p-index[i])/k[i] << ")\n";
				#endif
      v.push_back ( x[i]+(p-index[i])/k[i] );
    }
  }
  // The last region and upper boundary
  unsigned int lastRegionIndex = index[end];
				#ifdef TRACE_G0
				std::cout << "end " << end 
				          << " Last Region index " 
					  << index[end] << "\n";
				#endif
  for (unsigned int q = lastRegionIndex; q < lastIndex; ++q) {
				#ifdef TRACE_G0
				std::cout << "q " << q << " lastRegionIndex " 
					  << lastRegionIndex
        				  <<  " k[end] " << k[end] 
					  <<  " x[end] " << x[end] 
					  << " push_back(" 
					  << x[end]+(q-index[end])/k[end] 
					  << ")\n";
				#endif
    v.push_back ( x[end]+(q-lastRegionIndex)/k[end] );
  }
  v.push_back ( upperBoundary );
				#ifdef TRACE_G0
				std::cout << "upperBoundary push_back(" 
					  << upperBoundary << ")\n";
				#endif
  assert (v.size() == pointCount);
  return v;
}

std::ostream &
Grid<1>::put (std::ostream & os) const {
  if ( lastIndex == -1 ) {
    os << "Default-constructed Grid<1> \n\n";
    return os;
  }
  unsigned int N = index.size();
  std::vector<unsigned long> vx(2); 
  std::vector<unsigned long> vh(2); 
  std::vector<unsigned long> vk(2); 
  os << "Grid<1> with " << N << " regions:\n";
  for ( unsigned int i = 0; i < N; ++i ) {
    vx = DoubConv::dto2longs(x[i]);
    vh = DoubConv::dto2longs(h[i]);
    vk = DoubConv::dto2longs(k[i]);
    os  << i 
	<< "  index: " << index[i] 
	<< "  x: " << x[i] << " " << vx[0] << " " << vx[1] << "\n"
        <<"              h: " << h[i] << " " << vh[0] << " " << vh[1] << "\n"
      << "                 k: " << k[i] << " " << vk[0] << " " << vk[1] << "\n";
  }
  vx = DoubConv::dto2longs(penultimatePoint);
  vk = DoubConv::dto2longs(penultimateK);
  os    << " penultimatePoint: " << penultimatePoint << " "
        << vx[0] << " " << vx[1] << "\n"
	<< " penultimateK: " << penultimateK << " "
	<< vk[0] << " " << vk[1] <<"\n";
  vx = DoubConv::dto2longs(upperBoundary);
  os    << " lastIndex: " << lastIndex
  	<< "  upperBoundary: " << upperBoundary << " "
  	<< vx[0] << " " << vx[1] << "\n"
	<< " pointCount: " << pointCount << "\n";
}

std::istream &
Grid<1>::get (std::istream & is) {
  x.clear(); h.clear(); k.clear(); index.clear();
  std::string in;
  is >> in;
  if ( in == "Default-constructed" ) {
    is >> in;
    *this = Grid<1>();
    return is;
  }
  is >> in; // with
  unsigned int N;
  is >> N;
  is >> in; // regions:
  std::vector<unsigned long> vx(2); 
  std::vector<unsigned long> vh(2); 
  std::vector<unsigned long> vk(2); 
  double number;
  unsigned int iin;
  for ( unsigned int i = 0; i < N; ++i ) {
    is >> iin;
    assert (iin == i); // TODO -- replace this with a throw
    is >> in; // index:
    is >> iin;
    index.push_back(iin);
    is >> in; // x:
    is >> number; // x[i]
    is >> vx[0] >> vx[1]; number = DoubConv::longs2double(vx);
    x.push_back(number);
    is >> in; // h:
    is >> number; // h[i]
    is >> vh[0] >> vh[1]; number = DoubConv::longs2double(vh);
    h.push_back(number);
    is >> in; // k:
    is >> number; // k[i]
    is >> vk[0] >> vk[1]; number = DoubConv::longs2double(vk);
    k.push_back(number);
  }
  is >> in; // penultimatePoint:
  is >> number; // penultimatePoint
  is >> vx[0] >> vx[1]; penultimatePoint = DoubConv::longs2double(vx);
  is >> in; // penultimateK:
  is >> number; // penultimateK
  is >> vk[0] >> vk[1]; penultimateK = DoubConv::longs2double(vk);
  is >> in; // lastIndex:
  is >> lastIndex;
  is >> in; // upperBoundary:
  is >> number; // upperBoundary
  is >> vx[0] >> vx[1]; upperBoundary = DoubConv::longs2double(vx);
  is >> in; // pointCount:
  is >> pointCount;
  return is;
}

bool
Grid<1>::operator == (const Grid<1> & rhs) const {
  return    upperBoundary == rhs.upperBoundary 
         && penultimatePoint == rhs.penultimatePoint
	 && penultimateK == rhs.penultimateK
	 && lastIndex == rhs.lastIndex
	 && x == rhs.x
	 && h == rhs.h
	 && k == rhs.k
	 && index == rhs.index;
}

void
Grid<1>::shift(double delta) {
  upperBoundary += delta;
  penultimatePoint += delta;
  for (unsigned int i = 0; i < x.size(); ++i) {
    x[i] += delta;
  }
}


// =====================
// Multiple Grid Methods
// =====================

bool 
Grid<1>::search_and_push_back 
		(double p, std::vector<GridSearchAnswer> & answer) const {
  GridSearchAnswer gsa;
  bool res = search(p, gsa);
  answer.push_back(gsa);
  return res;
}



// =================
// Exception methods
// =================

const char* 
Grid_push_back_order_exception::what() const throw() {
  std::ostringstream os;
  os << "Grid<1>.push_back() for a point at " << a 
	<< " which is less than the upperBoundary " << b;
  return os.str().c_str();
}


const char* 
Grid_append_order_exception::what() const throw() {
  std::ostringstream os;
  os << "Grid<1> += a Grid<1> starting at " << a
	<< " which is less that the the upperBoundary " << b;
  return os.str().c_str();
}

const char* 
Grid_limits_order_exception::what() const throw() {
  std::ostringstream os;
  os << "ctor for Grid<1> requires upperLimit > lowerLimit, but "
     << a << " >= " << b;
  return os.str().c_str();
}

const char* 
Grid_face_dimension_exception::what() const throw() {
  std::ostringstream os;
  os << "Use of dimension " << n << " to specify a face of a Grid<" 
     << d << ">";
  return os.str().c_str();
}

} // end of namespace splines
